Require Import Vtac Program List Arith Even Div2 Relations String Max.
Require Import basis countable lang subs model basiclemmas weakiso transitivity_part1 transitivity_part2.
Require Import FunctionalExtensionality Classical ClassicalChoice IndefiniteDescription.

Set Implicit Arguments.

Local Notation cid := constructive_indefinite_description.

Section Bijective_Decomposition.

Variables
  (pD : list typename)
  (pG : list (termvar * typ))
  (pty : typ)
  (names: typename -> Prop) 
  (Nci : countably_infinite names).

Notation decomp := (@decomp pG pty _ Nci).
Notation namesnew := (@namesnew pG pty _ Nci).
Notation Ncinew := (@Ncinew pG pty _ Nci).

Inductive wsimilar sim (ty1 ty2 : typ) : Prop :=
 | wsim_closed (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2)
 | wsim_sim (SS : sim ty1 ty2)
 .

Inductive similar : typ -> typ -> Prop :=
 | sim_name : forall a (NIN : ~ namesnew a), similar (Tname a) (Tname a)
 | sim_bvar : forall x, similar (Tbvar x) (Tbvar x)
 | sim_unit : similar Tunit Tunit
 | sim_int : similar Tint Tint
 | sim_bool : similar Tbool Tbool
 | sim_prod : forall ty1 ty2 (WS : wsimilar similar ty1 ty2)
                      ty1' ty2' (WS' : wsimilar similar ty1' ty2'),
               similar (Tprod ty1 ty1') (Tprod ty2 ty2')
 | sim_sum : forall ty1 ty2 (WS : wsimilar similar ty1 ty2)
                     ty1' ty2' (WS' : wsimilar similar ty1' ty2'),
              similar (Tsum ty1 ty1') (Tsum ty2 ty2')
 | sim_arrow : forall ty1 ty2 (WS : wsimilar similar ty1 ty2)
                       ty1' ty2' (WS' : wsimilar similar ty1' ty2'),
                similar (Tarrow ty1 ty1') (Tarrow ty2 ty2')
 | sim_mu : forall ty1 ty2 (WS : wsimilar similar ty1 ty2),
             similar (Tmu ty1) (Tmu ty2)
 | sim_exists : forall ty1 ty2 (WS : wsimilar similar ty1 ty2),
                 similar (Texists ty1) (Texists ty2)
 | sim_forall : forall ty1 ty2 (WS : wsimilar similar ty1 ty2),
                 similar (Tforall ty1) (Tforall ty2)
 | sim_ref : forall ty1 ty2 (WS : wsimilar similar ty1 ty2),
              similar (Tref ty1) (Tref ty2)
 .
Hint Constructors similar.

Definition not_similar_set
  := { pty : typ * typ | lc_typ 0 (fst pty) /\ lc_typ 0 (snd pty) /\ ~ similar (fst pty) (snd pty) }.

Lemma not_similar_set_cntinf: countably_infinite_set not_similar_set.
Proof.
  assert (X:= cntinf_prod typ_cntinf typ_cntinf). 
  eapply countably_infinite_set_impl_set2 in X; red in X; des.
  eapply countably_infinite_set2_impl_set.
  exists (fun pty : not_similar_set => f (`pty)).
  assert (pf: forall n, True /\ True /\ ~ similar (Tname n) Tunit).
    by intros; do 2 (split; eauto); intro NS; inv NS.
  exists (fun n => exist _ (Tname n, Tunit) (pf n) : not_similar_set).
  split; red; intros.
  - by eapply X, proj1_sig_inj in H.
  - by inv H.
Qed.

Lemma namesnew_not_similar_bij: 
  exists fg : (not_similar_set -> sig namesnew) * (sig namesnew -> not_similar_set),
    (forall a, (fst fg) ((snd fg) a) = a) /\
    (forall pty, (snd fg) ((fst fg) pty) = pty).
Proof.
  assert (X:= not_similar_set_cntinf).
  assert (Y:= Ncinew).
  red in X; rrdes Y.
  exists ((fun pty => g (f0 pty)), (fun a => g0 (f a))).
  by simpl; split; intros; [rewrite X0, Y | rewrite Y0, X].
Qed.

Definition TEnc := fst (proj1_sig (cid _ namesnew_not_similar_bij)).

Definition TEncInv := snd (proj1_sig (cid _ namesnew_not_similar_bij)).

Lemma TEnc_TEncInv: forall a, TEnc (TEncInv a) = a.
Proof. apply (proj2_sig (cid _ namesnew_not_similar_bij)). Qed.

Lemma TEncInv_TEnc: forall pty, TEncInv (TEnc pty) = pty.
Proof. apply (proj2_sig (cid _ namesnew_not_similar_bij)). Qed.

Lemma TEnc_inj : Image.injective _ _ TEnc.
Proof. by red; intros; rewrite <- TEncInv_TEnc, <- H, TEncInv_TEnc. Qed.

Lemma TEncInv_inj : Image.injective _ _ TEncInv.
Proof. by red; intros; rewrite <- TEnc_TEncInv, <- H, TEnc_TEncInv. Qed.

Definition nsfst (pty : not_similar_set) : typ := fst (`pty).

Definition nssnd (pty : not_similar_set) : typ := snd (`pty).

Notation nl := (fun proj a =>
   match LEM (namesnew a) with left pf => proj (TEncInv (exist _ a pf)) | _ => Tname a end).

Lemma nsfst_nssnd_inj: forall ty ty', nsfst ty = nsfst ty' -> nssnd ty = nssnd ty' -> ty = ty'.
Proof. 
  by intros; destructs ty ty' x x0; apply proj1_sig_inj; unfold nsfst, nssnd in *; simpls; subst.
Qed.

Lemma decomp_wsim : forall ty, wsimilar similar (decomp (nl nsfst) ty) (decomp (nl nssnd) ty).
Proof.
  induction ty; intros; simpl; try auto using wsim_sim.
  case (LEM (namesnew aa)); intros.
    by apply wsim_closed; apply (proj2_sig (TEncInv _)).
  by apply wsim_sim; auto.
Qed.

Lemma decomp_sim1 : forall ty (SS : similar (decomp (nl nsfst) ty) (decomp (nl nssnd) ty)),
  ~ exists a, ty = Tname a /\ namesnew a.
Proof.
  intro; case ty; intros; simpl; auto; try ( intro X; destruct X as (? & A & B); inversion A).
  revert SS; simpl; case (LEM (namesnew aa)); intros.
    by eapply (proj2_sig (TEncInv _)); eauto.
  by inversion SS; rewrite H0 in n; contradiction.
Qed.

Lemma decomp_sim2 : forall ty (NN : ~ exists a, ty = Tname a /\ namesnew a),
  similar (decomp (nl nsfst) ty) (decomp (nl nssnd) ty).
Proof.
  intro; case ty; intros; simpl; eauto using decomp_wsim.
  by case (LEM (namesnew aa)); intros; auto; contradict NN; exists aa; auto.
Qed.

Lemma decomp_sim : forall ty,
  (~ exists a, ty = Tname a /\ namesnew a) <-> similar (decomp (nl nsfst) ty) (decomp (nl nssnd) ty).
Proof.
  by split; auto using decomp_sim1, decomp_sim2.
Qed.

Lemma wsim_decomp_aux : forall ty1 ty2 (NSS : ~ similar ty1 ty2)
    (WS : wsimilar similar ty1 ty2),
  exists ty, decomp (nl nsfst) ty = ty1 /\ decomp (nl nssnd) ty = ty2.
Proof.
  intros; destruct WS; try contradiction.
  exists (Tname (proj1_sig (TEnc (exist _ (ty1,ty2) (conj CL1 (conj CL2 NSS)))))).
  simpl; case (LEM (namesnew (proj1_sig (TEnc (exist _ (ty1,ty2) (conj CL1 (conj CL2 NSS))))))).
    by intros; rewrite sig_eta, TEncInv_TEnc; simpls.
  by intros; contradict n; eapply (proj2_sig (TEnc _)).
Qed.

Lemma wsim_decomp : forall ty1 ty2 (WS : wsimilar similar ty1 ty2),
  exists ty, decomp (nl nsfst) ty = ty1 /\ decomp (nl nssnd) ty = ty2.
Proof.
  induction ty1; intros.
  - by destruct (classic (similar (Tname aa) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux];
       inversion SS; exists (Tname aa); simpl; case (LEM (namesnew aa)); [contradiction |].
  - by destruct (classic (similar (Tbvar bb) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux]; 
      inversion SS; exists (Tbvar bb).
  - by destruct (classic (similar Tunit ty2)) as [SS|SS]; [| by apply wsim_decomp_aux];
      inversion SS; exists Tunit.
  - by destruct (classic (similar Tint ty2)) as [SS|SS]; [| by apply wsim_decomp_aux];
      inversion SS; exists Tint.
  - by destruct (classic (similar Tbool ty2)) as [SS|SS]; [| by apply wsim_decomp_aux];
      inversion SS; exists Tbool.
  - destruct (classic (similar (Tprod ty1_1 ty1_2) ty2)) as [SS|SS];
      [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1_1; eauto; edestruct IHty1_2; eauto.
    exists (Tprod x x0); simpl; destruct H as (E1f, E1s); destruct H3 as (E2f, E2s).
    by rewrite E1f, E1s, E2f, E2s.
  - destruct (classic (similar (Tsum ty1_1 ty1_2) ty2)) as [SS|SS];
      [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1_1; eauto; edestruct IHty1_2; eauto.
    exists (Tsum x x0); simpl; destruct H as (E1f, E1s); destruct H3 as (E2f, E2s).
    by rewrite E1f, E1s, E2f, E2s.
  - destruct (classic (similar (Tarrow ty1_1 ty1_2) ty2)) as [SS|SS];
      [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1_1; eauto; edestruct IHty1_2; eauto.
    exists (Tarrow x x0); simpl; destruct H as (E1f, E1s); destruct H3 as (E2f, E2s).
    by rewrite E1f, E1s, E2f, E2s.
  - destruct (classic (similar (Tref ty1) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1; eauto.
    by exists (Tref x); simpl; destruct H as (E1f, E1s); rewrite E1f, E1s.
  - destruct (classic (similar (Tmu ty1) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1; eauto.
    by exists (Tmu x); simpl; destruct H as (E1f, E1s); rewrite E1f, E1s.
  - destruct (classic (similar (Tforall ty1) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1; eauto.
    by exists (Tforall x); simpl; destruct H as (E1f, E1s); rewrite E1f, E1s.
  - destruct (classic (similar (Texists ty1) ty2)) as [SS|SS]; [| by apply wsim_decomp_aux].
    inversion SS; edestruct IHty1; eauto.
    by exists (Texists x); simpl; destruct H as (E1f, E1s); rewrite E1f, E1s.
Qed.

Definition nfst := nl nsfst.

Definition nsnd := nl nssnd.

Lemma nfst_closed : forall a, lc_typ 0 (nfst a).
Proof.
  intros; unfold nfst; simpl; case (LEM (namesnew a)); intros; simpl; eauto.
  generalize (TEncInv (exist namesnew a n)).
  by intros; unfold nsfst; destruct n0; simpl; des; eauto.
Qed.

Lemma nsnd_closed : forall a, lc_typ 0 (nsnd a).
Proof.
  intros; unfold nsnd; simpl; case (LEM (namesnew a)); intros; simpl; eauto.
  generalize (TEncInv (exist namesnew a n)).
  by intros; unfold nsfst; destruct n0; simpl; des; eauto.
Qed.

Lemma decomp_inj : forall ty ty'
  (EQ1 : decomp nfst ty = decomp nfst ty') (EQ2 : decomp nsnd ty = decomp nsnd ty'),
  ty = ty'.
Proof.
  unfold nfst, nsnd; induction ty; intros;
    try (
      by generalize dependent ty'; intro ty'; case ty'; intros; simpl in EQ1, EQ2;
      try inversion EQ1; auto; case (LEM (namesnew aa)) in EQ1, EQ2; [| by inversion EQ1];
      assert (~ similar (nsfst (TEncInv (exist _ aa n))) (nssnd (TEncInv (exist _ aa n)))); 
      [by eapply (proj2_sig (TEncInv _)); eauto |];
      contradict H; rewrite <- EQ1, <- EQ2; auto
    );
    try (
      by assert (NN : ~ exists a, ty' = Tname a /\ namesnew a);
        [by apply decomp_sim; simpl in EQ1, EQ2; rewrite <- EQ1, <- EQ2; auto using decomp_wsim |];
      generalize dependent ty'; intro ty'; case ty'; intros; simpl in EQ1, EQ2; inversion EQ1;
        [by revert EQ1 EQ2; simpl; case (LEM (namesnew aa)); intros; inversion EQ1; contradict NN;
          exists aa; auto |];
      f_equal; inversion EQ2; auto using IHty1, IHty2
    );
    try (
      by assert (NN : ~ exists a, ty' = Tname a /\ namesnew a);
        [by apply decomp_sim; simpl in EQ1, EQ2; rewrite <- EQ1, <- EQ2; auto using decomp_wsim |];
      generalize dependent ty'; intro ty'; case ty'; intros; simpl in EQ1, EQ2; inversion EQ1;
        [by revert EQ1 EQ2; simpl; case (LEM (namesnew aa)); intros; inversion EQ1; contradict NN;
          exists aa; auto |];
      f_equal; inversion EQ2; auto using IHty
    ).

  Case "name".
  revert EQ1 EQ2; simpl; case (LEM (namesnew aa)); intros.
    assert (NSS : ~ similar (Tname aa) ty').
      by intro A; inversion A; contradiction.
    assert (N : exists a, ty' = Tname a /\ namesnew a).
      apply NNPP; intro; eapply (proj2_sig (TEncInv (exist _ aa n))).
      by unfold nsfst, nssnd in *; rewrite EQ1, EQ2; eapply decomp_sim.
    destruct N as (a' & E & N); rewrite E; f_equal; subst. 
    change aa with (proj1_sig (exist _ aa n)); change a' with (proj1_sig (exist _ a' N)).
    setoid_rewrite <- TEnc_TEncInv; simpls; revert EQ1 EQ2.
    case (LEM (namesnew a')); intros; try contradiction.
    by f_equal; f_equal; replace N with n0; eauto using proof_irr, nsfst_nssnd_inj. 
  revert EQ1 EQ2; case ty'; simpl; intros; inversion EQ1; revert EQ1 EQ2 H0.
  case (LEM (namesnew aa0)); intros; auto; contradict n; apply NNPP; intro.
  assert (similar (decomp (nl nsfst) (Tname aa)) (decomp (nl nssnd) (Tname aa))).
    apply decomp_sim; intro; destruct H1 as []; destruct H1; inversion H1; rewrite <- H4 in H2.
    by contradiction.
  simpl in H1; case (LEM (namesnew aa)) in H1; try contradiction.
  rewrite EQ1 in H1 at 1; rewrite EQ2 in H1; contradict H1. 
  by eapply (proj2_sig (TEncInv _)); eauto.
Qed.

Lemma decomp_surj : forall ty1 ty2 (CL1 : lc_typ 0 ty1) (CL2 : lc_typ 0 ty2),
  exists ty, decomp nfst ty = ty1 /\ decomp nsnd ty = ty2.
Proof.
  by intros; apply wsim_decomp; apply wsim_closed.
Qed.

End Bijective_Decomposition.

Theorem transitivity : forall D G ty e1 e2 e3
  (WR1 : wreq D G ty e1 e2) (WR2 : wreq D G ty e2 e3),
  wreq D G ty e1 e3.
Proof.
  intros; intros names Nci.
  specialize (WR1 _ (Nci1 Nci)); des.
  specialize (WR2 _ (Nci2 G ty Nci)); des.
  assert (X: forall a w (NA: wnam (wlift w) a), wnam w a).
    by intros; rrdes NA; [contradiction NA | eauto].
  exists (lpw G ty Nci w w0 (nfst G ty Nci) (nsnd G ty Nci) 
             (nfst_closed G ty Nci) (nsnd_closed G ty Nci) 
             (decomp_inj G ty Nci) (decomp_surj G ty Nci)).
  split; try split; unnw;
  [eapply lpw_names | eapply lpw_stable |]; eauto.
  by rdes WEQ; rdes WEQ0; eapply weq_iso; [eapply phipsi_is_wkiso | eapply weq_trans];
     eauto using decomp_inj, decomp_surj.
Qed.

